package org.securedroneid.android.bluetooth;

import static org.securedroneid.android.Constants.DET_SIZE;
import static org.securedroneid.android.Constants.HASH_SIZE;
import static org.securedroneid.android.Constants.SIGNATURE_SIZE;

import android.util.Log;

import org.bouncycastle.math.ec.rfc8032.Ed25519;
import org.bouncycastle.util.encoders.Hex;
import org.securedroneid.android.Constants;
import org.securedroneid.android.data.AircraftObject;
import org.securedroneid.android.data.AuthenticationData;
import org.securedroneid.android.data.Connection;
import org.securedroneid.android.data.Identification;
import org.securedroneid.android.data.Link;
import org.securedroneid.android.data.LocationData;
import org.securedroneid.android.data.Manifest;
import org.securedroneid.android.data.OperatorIdData;
import org.securedroneid.android.data.SelfIdData;
import org.securedroneid.android.data.SystemData;
import org.securedroneid.android.data.Wrapper;

import java.io.StringWriter;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

public class DRIPDataManager{

    static Map<String, byte []> hashes = new HashMap<String, byte[]>();

    static Set<DRIPLinkObject> pendingLinks= new HashSet<>();

    //This should be the DET and HI(public key) for the entities in the highest level of endorsement chain
    //So far we don't know the Public key of the RAA(1026) so this is set to Auth HDA
    //please see RFC9575 section 6.3 and draft-drip-dki-3 section 1 for more detail
    //TODO put these into separate files to read...
    static String HEX_Auth_HDA_DET = "2001003ffe0402050637ac6f167a2303";

    static byte[] Auth_HDA_PUB_KEY = new byte[] {
            (byte) 0xAE, (byte) 0x16, (byte) 0x85, (byte) 0x4F,
            (byte) 0x26, (byte) 0xC1, (byte) 0xB9, (byte) 0x98,
            (byte) 0x0C, (byte) 0xE4, (byte) 0x49, (byte) 0x35,
            (byte) 0x4D, (byte) 0xDF, (byte) 0xE8, (byte) 0x3E,
            (byte) 0x25, (byte) 0x8E, (byte) 0x94, (byte) 0x8F,
            (byte) 0x60, (byte) 0xD7, (byte) 0x6B, (byte) 0x11,
            (byte) 0x14, (byte) 0x5D, (byte) 0x72, (byte) 0xDA,
            (byte) 0x1C, (byte) 0x34, (byte) 0x34, (byte) 0x29
    };

    static boolean verified = false;

    static String TAG = "DRIPDataManager";

    public enum SAMType {
        None(0),
        Link(1),
        Wrapper(2),
        Manifest(3),
        Frame(4);

        SAMType(int id) { this.id = id; }
        public final int id;

        public static SAMType fromId(int id) {
            if (id == None.id) {
                return None;
            } else if (id == Link.id) {
                return Link;
            } else if (id == Wrapper.id) {
                return Wrapper;
            } else if (id == Manifest.id) {
                return Manifest;
            } else if (id == Frame.id) {
                return Frame;
            } else {
                return null;
            }
        }
    }

    public static void handleSAMMessage(ByteBuffer SAMMessageData, AircraftObject ac, AuthenticationData auth, OpenDroneIdDataManager dataManager) {

        Log.d(TAG,"Started handling SAM");

        if(SAMMessageData.remaining() == 0) {
            Log.e(TAG, "SAM Message is empty");
            return;
        }

        byte type_int = SAMMessageData.get();
        SAMType type = SAMType.fromId(type_int);

        if(type == null) {
            Log.e(TAG, "SAM Type is null (" + type_int + ")");
            return;
        }

        switch(type) {
            case Link:
            handleLink(SAMMessageData, auth, ac);
            break;
            case Wrapper:
            case Manifest:
            handleEndorsementStructure(SAMMessageData, type, ac, auth, dataManager);
            break;
            default:
        }
    }
    private static void recursivelyProcessingLinks(DRIPLinkObject verifiedLink,AircraftObject ac,AuthenticationData auth) {
        if(ac.isDET(verifiedLink.childDET)){
            Log.d(TAG,"All link message has been verified");
            //Match BE(UA,HDA) and verified
            Link link = new Link();
            link.setPublicKey(verifiedLink.parentHI);
            link.setMsgCounter(auth.getMsgCounter());
            link.setTimestamp(auth.getTimestamp());
            ac.link.postValue(link);
            ac.setPublicKey(verifiedLink.childHI);
            return;
        }

        Iterator<DRIPLinkObject> iterator = pendingLinks.iterator();
        while (iterator.hasNext()) {
            DRIPLinkObject pendingLink = iterator.next();
            if (Arrays.equals(pendingLink.parentDET, verifiedLink.childDET)) {
                // Verify the pending link with the parent's public key
                pendingLink.verifySignature(verifiedLink.childHI);
                if (pendingLink.isVerified) {

                    iterator.remove();
                    recursivelyProcessingLinks(pendingLink,ac,auth);
                }
            }
        }
    }

    private static void handleLink(ByteBuffer linkData, AuthenticationData auth, AircraftObject ac){
        //Follow RFC9575 6.3
        DRIPLinkObject dripLinkObject = new DRIPLinkObject(linkData);

        if(Arrays.equals(dripLinkObject.parentDET,Hex.decode(HEX_Auth_HDA_DET))){
            // This should be endorsement from highest level of endorsement chain
            // Current is BE(issue_HDA, auth_HDA) endorsement
            dripLinkObject.verifySignature(Auth_HDA_PUB_KEY);
            if(dripLinkObject.isVerified){
                // Recursively Handle the pending drip link message
                recursivelyProcessingLinks(dripLinkObject,ac,auth);
            }
        }
        else{
            pendingLinks.add(dripLinkObject);

        }
    }


    private static void handleEndorsementStructure(ByteBuffer endorsementStructureData, SAMType type, AircraftObject ac, AuthenticationData auth, OpenDroneIdDataManager dataManager) {
        int messageSize = endorsementStructureData.remaining() - SIGNATURE_SIZE;
        Log.d(TAG,"Started handling endorsement structure (" + messageSize + ")");
        byte[] entire_message = new byte[messageSize];
        endorsementStructureData.slice().get(entire_message);

        int VNB = endorsementStructureData.getInt();
        int VNA = endorsementStructureData.getInt();

        if(!verifyTimestamps(VNB, VNA)) {
            Log.e(TAG, "Timestamps were not verified successfully");
            return;
        }

        int evidenceSize = endorsementStructureData.remaining() - SIGNATURE_SIZE - DET_SIZE;
        ByteBuffer evidence = endorsementStructureData.slice();
        evidence.order(ByteOrder.LITTLE_ENDIAN).limit(evidenceSize);

        // Advance past the evidence
        endorsementStructureData.position(endorsementStructureData.position() + evidenceSize);

        byte[] DET = new byte[DET_SIZE];
        endorsementStructureData.get(DET);

        byte[] publicKey = ac.getPublicKey();

        if(!ac.isDET(DET) || publicKey == null) {
            Log.e(TAG, "The DET is incorrect or the public key is still null");
            return;
        }

        byte[] signature = new byte[Constants.SIGNATURE_SIZE];
        endorsementStructureData.get(signature);

        boolean signatureValid = Ed25519.verify(
                signature, 0, publicKey, 0, entire_message, 0, messageSize);

        if(!signatureValid) {
            Log.e(TAG, "The signature verification was not successful");
            StringWriter sw = new StringWriter();
            for(byte b : entire_message) {
                sw.append(String.format("%02X ", b));
            }
            Log.d(TAG, "Entire message: " + sw);
            return;
        }

        switch (type) {
            case Wrapper:
                handleWrapper(evidence, auth, dataManager, ac);
                break;
            case Manifest:
                handleManifest(evidence, auth, ac);
                break;
            default:
        }
    }

    private static void handleWrapper(ByteBuffer wrapperData, AuthenticationData auth, OpenDroneIdDataManager dataManager, AircraftObject ac) {
        if ((wrapperData.remaining() % 25) != 0) {
            Log.e(TAG, "Wrapper sanity check failed");
            return;
        }
        int messageCount = wrapperData.remaining() / 25;

        Log.d(TAG,"Started handling wrapper (" + messageCount + ")");

        Connection connection = ac.connection.getValue();
        if(connection == null) {
            Log.e(TAG, "Connection is null");
            return;
        }

        OpenDroneIdParser.Type[] messageTypes = new OpenDroneIdParser.Type[messageCount];

        for(int i = 0; i < messageCount; i++) {
            byte[] message = new byte[25];
            wrapperData.get(message);
            messageTypes[i] = dataManager.receiveVerifiedData(message, auth.getTimestamp(), connection.macAddress, ac.getMacAddress(), connection.rssi, connection.transportType);
        }

        Wrapper wrapper = new Wrapper(messageCount);
        wrapper.setMessageTypes(messageTypes);
        wrapper.setTimestamp(auth.getTimestamp());
        wrapper.setMsgCounter(auth.getMsgCounter());
        ac.wrapper.postValue(wrapper);
    }

    private static void handleManifest(ByteBuffer manifestData, AuthenticationData auth, AircraftObject ac) {
        int hashLength = 8;
        if ((manifestData.remaining() % hashLength) != 0) {
            return;
        }
        int hashCount = (manifestData.remaining() / hashLength) - 3;

        // TODO: Check previous hash and store current hash

        byte[] previousHash = new byte[Constants.HASH_SIZE];
        manifestData.get(previousHash);

        byte[] currentHash = new byte[Constants.HASH_SIZE];
        manifestData.get(currentHash);

        Identification id2 = ac.identification2.getValue();


        if (hashes.containsKey(id2.getUasIdAsString())){
            if (Arrays.equals(previousHash, hashes.get(id2.getUasIdAsString()))) {
                Log.i(TAG, "id2 UAS ID " + id2.getUasIdAsString());
                verified = true;
            } else {
                verified = false;
                Log.i(TAG, "False manifest");
            }
        }
        Log.i(TAG, "");
        hashes.put(id2.getUasIdAsString(), currentHash);


        Log.i(TAG, "prev " + Arrays.toString(previousHash)); //

        for (Map.Entry<String, byte[]> entry : hashes.entrySet()) {
            Log.i(TAG, "Key: " + entry.getKey() + ", Value: " + Arrays.toString(entry.getValue()));}

        if (verified) {
            byte[] dripLinkHash = new byte[HASH_SIZE];
            manifestData.get(dripLinkHash);
            // TODO: Verify the hash of the link 

            Log.i(TAG, "Is verified, posting data.");
            ArrayList<OpenDroneIdParser.Type> messageTypes = new ArrayList<>();

            for (int i = 0; i < hashCount; i++) {
                byte[] messageHash = new byte[HASH_SIZE];
                manifestData.get(messageHash);

                // There is probably a better way of doing this by changing the way the messages are stored

                Identification identification1 = ac.identification1.getValue();
                if (identification1 != null && !identification1.isVerified()) {
                    if (Arrays.equals(identification1.getHash(), messageHash)) {
                        identification1.setVerified(true);
                        ac.identification1.postValue(identification1);
                        messageTypes.add(OpenDroneIdParser.Type.BASIC_ID);
                    }
                }

                Identification identification2 = ac.identification2.getValue();
                Log.i(TAG, "id" + identification2.toString());
                if (identification2 != null && !identification2.isVerified()) {
                    if (Arrays.equals(identification2.getHash(), messageHash)) {
                        identification2.setVerified(true);
                        ac.identification2.postValue(identification2);
                        messageTypes.add(OpenDroneIdParser.Type.BASIC_ID);
                    }
                }

                LocationData location = ac.location.getValue();
                if (location != null && !location.isVerified()) {
                    if (Arrays.equals(location.getHash(), messageHash)) {
                        location.setVerified(true);
                        ac.location.postValue(location);
                        messageTypes.add(OpenDroneIdParser.Type.LOCATION);
                    }
                }

                SelfIdData selfid = ac.selfid.getValue();
                if (selfid != null && !selfid.isVerified()) {
                    if (Arrays.equals(selfid.getHash(), messageHash)) {
                        selfid.setVerified(true);
                        ac.selfid.postValue(selfid);
                        messageTypes.add(OpenDroneIdParser.Type.SELFID);
                    }
                }

                SystemData system = ac.system.getValue();
                if (system != null && !system.isVerified()) {
                    if (Arrays.equals(system.getHash(), messageHash)) {
                        system.setVerified(true);

                        ac.system.postValue(system);
                        messageTypes.add(OpenDroneIdParser.Type.SYSTEM);
                    }
                }

                OperatorIdData operatorid = ac.operatorid.getValue();
                if (operatorid != null && !operatorid.isVerified()) {
                    if (Arrays.equals(operatorid.getHash(), messageHash)) {
                        operatorid.setVerified(true);
                        ac.operatorid.postValue(operatorid);
                        messageTypes.add(OpenDroneIdParser.Type.OPERATOR_ID);
                    }
                }
            }

            Manifest manifest = new Manifest();
            manifest.setMessageTypes(messageTypes.toArray(new OpenDroneIdParser.Type[0]));
            manifest.setTimestamp(auth.getTimestamp());
            manifest.setMsgCounter(auth.getMsgCounter());
            Log.i(TAG, "manifest: " + manifest.toString());
            ac.manifest.postValue(manifest);
        }

    }

    private static boolean verifyTimestamps(int ValidNotBeforeTimestamp, int ValidNotAfterTimestamp) {
        Timestamp VNB = new Timestamp((Constants.TIMESTAMP_OFFSET + ValidNotBeforeTimestamp) * 1000L);
        Timestamp VNA = new Timestamp((Constants.TIMESTAMP_OFFSET + ValidNotAfterTimestamp) * 1000L);

        Timestamp now = new Timestamp(System.currentTimeMillis());

        if(VNB.after(now)) {
            Log.e(TAG, "VNB timestamp is after current time");
            return false;
        }
        if(VNA.before(now)) {
            Log.e(TAG, "VNA timestamp (" + VNA + ") is already expired");
            return false;
        }
        return true;
    }

}